{-# LANGUAGE BangPatterns #-}
{-| Module      : Gargantext.Core.Viz.Graph.ProxemyOptim
Description : Proxemy
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

(/!\ To be published soon, confidential document for now. After
publication it will be integrated to the backend with usual license --
see above)

Article: Confluence for Graph Clustering, B. Gaume and A. Delanoë, Journal:TBA

Code written in Haskell by A. Delanoë from first Python Specifications by B.
Gaume.

-}

{-# LANGUAGE FlexibleContexts
           , DataKinds
           , GADTs
           , KindSignatures
           , ScopedTypeVariables
           , StandaloneDeriving
           , TypeOperators
           , TypeApplications
           , NoImplicitPrelude
           , RankNTypes
           , MultiParamTypeClasses
#-}

{-# LANGUAGE AllowAmbiguousTypes #-}
module Graph.BAC.ProxemyOptim
  where

import Data.IntMap (IntMap)
import Data.Maybe (isJust)
import Data.Proxy (Proxy(Proxy))
import Data.Reflection
import GHC.TypeLits (KnownNat, Nat, SomeNat(SomeNat), type(+), natVal, sameNat, someNatVal)
import Graph.FGL
import Graph.Types
import Prelude (String, readLn, error)
import Protolude hiding (sum, natVal, trace)
import qualified Data.Graph.Inductive              as DGI
import qualified Data.Graph.Inductive.PatriciaTree as DGIP
import qualified Data.List                         as List
import qualified Data.IntMap                       as Dict
import qualified Data.Vector                       as V
import qualified Data.Vector.Storable              as VS
import qualified Prelude                           as Prelude
import qualified Data.Set                          as Set

import qualified Data.Matrix.Sparse.Static         as SMatrix
import qualified Data.Vector.Sparse.Static         as SV
import qualified Numeric.LinearAlgebra.Static      as DMatrix
import Debug.Trace
import qualified Data.Vector.Unboxed as UV

----------------------------------------------------------------


clusteringOptim :: forall n a b. (KnownNat n, Show a, Show b)
                => Length
                -> Similarity
                -> FiniteGraph n a b
                -> Clustering Node
clusteringOptim l s fg@(FiniteGraph g) = trace "clusteringOptim" $ -- traceShow ("mc", mc) $
  make_clust_part sorted_edges matq
    where
      adj    = trace "adjacent False" $ adjacent fg False
      tra    = trace "tra" $ transition adj

      adj'   = trace "adjacent True" $ adjacent fg True
      tra'    = trace "tra'" $ transition adj'
      sorted_edges = adj `seq` tra `seq` adj' `seq` tra' `seq`
                     (trace "sorted_edges"
                   $ sort_edges (natToInt @n)
                   $ edges_confluence l adj' tra')

      mc    = matconf False adj (proxemie l tra)
      md    = matmod fg

      matq = case s of
        Conf -> SimConf mc --  $ matconf False adj (proxemie l tra)
        Mod  -> SimMod  md --  $ matmod fg

-- | Where main Types are defined as
data Similarity = Conf | Mod
type Length      = Int

-- | A finite Graph is a Graph whose number of Nodes is a KnownNat
data FiniteGraph (n :: Nat) a b = FiniteGraph (Graph a b)

data Clustering a = ClusteringIs { parts :: !(Dict (Set a))
                                 , index :: !(Dict Int)
                                 , score :: !(Double)
                                 , mode  :: !ClusteringMode
                                 }
                  | Clusterings { strict :: !(StrictClustering   a)
                                , over   :: !(OverlapsClustering a)
                                }
  deriving (Show, Eq)

data ClusteringMode = Part | Over | Both
  deriving (Show, Eq)

type StrictClustering   a = Clustering a
type OverlapsClustering a = Clustering a

type IsReflexive = Bool
type NeighborsFilter a b = DGI.Gr a b -> Node -> [Node]
type RmEdge      = Bool
---------------------------------------------------------------
-- Data Structure
type VectorS          n = SMatrix.Matrix 1 n Double
type VectorD          n = DMatrix.L 1 n

type AdjacencyMatrix  n = MatrixS n
type TransitionMatrix n = MatrixS n
type ProxemyMatrix    n = MatrixD n

type ConfluenceMatrix n = MatrixD n
type ModularityMatrix n = MatrixD n

data SimilarityMatrix n = SimConf !(ConfluenceMatrix n)
                        | SimMod  !(ModularityMatrix n)


-----
adjacent :: KnownNat n
         => FiniteGraph n a b
         -> IsReflexive
         -> AdjacencyMatrix n
adjacent (FiniteGraph g) isReflexive = trace "adjacent"
  SMatrix.fromList $ triplets <> diag
  where
    triplets = [ (i, j, 1.0)
               | i <- nodes g
               , j <- neighbors g i
               , i /= j
               ]

    diag = case isReflexive of
      True  -> [ (n, n, 1.0) | n <- nodes g ]
      False -> []


transition :: (HasCallStack, KnownNat n)
           => AdjacencyMatrix  n
           -> TransitionMatrix n
transition m = trace "transition" $ SMatrix.imap (\i j v -> v * (SV.!) s i) m
  where
    s = sumWith Colonne (\s -> 1 / s) m


proxemie :: KnownNat         n
         => Length
         -> TransitionMatrix n
         -> ProxemyMatrix    n
proxemie l tm = trace ("proxemie (l = " ++ show l ++ ")") $ case l <= 1 of
  True  -> SMatrix.densify tm
  False -> matInfo "proxemie tm" tm $
    case matPow tm (l-1) of
      tm' -> matInfo "proxemie tm'" tm' $ SMatrix.densify tm'


-- https://en.wikipedia.org/wiki/Exponentiation_by_squaring
matPow :: MatrixS n -> Int -> MatrixS n
matPow mat n
  | n <= 0 = error "matPow: positive exponent expected"
  | otherwise = f mat n
  where
    f a i
      | even i = f (a `mul` a) (i `quot` 2)
      | i == 1 = a
      | otherwise = g (a `mul` a) (i `quot` 2) a
    g a i b
      | even i = g (a `mul` a) (i `quot` 2) b
      | i == 1 = a `mul` b
      | otherwise = g (a `mul` a) (i `quot` 2) (a `mul` b)
    mul = SMatrix.mul

---------------------------------------------------------------
matconf :: forall n. KnownNat n
        => RmEdge
        -> AdjacencyMatrix  n
        -> ProxemyMatrix    n
        -> ConfluenceMatrix n
matconf False a p = seq confmat $ trace "matconf" $ symmetry confmat
  where
    -- vcount  = natToInt @n
    degs    = DMatrix.extract $ SV.toDense (sumWith Colonne identity a)
    sumdeg  = VS.sum degs
    tp      = DMatrix.tr p
    confmat = seq tp . trace "confmat" $
      DMatrix.imapL
        (\(x, y) v ->
            if x < y
            then let prox_y_x_length = v
                     prox_y_x_infini = ((VS.!) degs x) / sumdeg
                 in
                   (prox_y_x_length - prox_y_x_infini)
                 / (prox_y_x_length + prox_y_x_infini)
            else 0
        ) tp
matconf True _a _p = panic "MatConf True: TODO but not needed for now"


matmod :: forall n a b. KnownNat n => FiniteGraph n a b -> ModularityMatrix n
matmod fg = trace "matmod" $ symmetry $ SMatrix.densify modmat
  where
    n' = natToInt @n
    a  = adjacent fg False
    !sumRows = DMatrix.extract $ SV.toDense (sumWith Ligne   identity a)
    !sumCols = DMatrix.extract $ SV.toDense (sumWith Colonne identity a)
    !ecount = SMatrix.sum a
    modmat = SMatrix.imap (\x y v ->
                                if x < y
                                then v - ((VS.!) sumRows x * (VS.!) sumCols y) / (2 * ecount)
                                else 0
                          ) a

---------------------------------------------------------------
type UnsortedEdges = [(Node, Node, Double)]
type SortedEdges   = [(Node, Node, Double)]

matInfo
  :: forall b (n :: Nat) (m :: Nat).
     (KnownNat n, KnownNat m)
  => String
  -> SMatrix.Matrix n m Double
  -> b -> b
matInfo name mat = trace $
  name ++ ": " ++ show nnz ++ "/" ++ show total ++
  " (~" ++ pcstr ++ "%) nnz, " ++
  szInfo ++
  "\n\t= [ " ++ intercalate ", " snip ++ " , ... ]"
  where total = n*m
        nnz = SMatrix.nonZeros mat
        pc = fromIntegral (100 * nnz) / fromIntegral total
        pcstr = take 4 (show pc)
        n = natToInt @n
        m = natToInt @m
        szInfo = show n ++ "x" ++ show m
        snip = [ show (coeffSM 0 i mat) | i <- [1..10] ]

edges_confluence :: forall           n
                  . KnownNat         n
                 => Length
                 -> AdjacencyMatrix  n
                 -> TransitionMatrix n
                 -> UnsortedEdges
edges_confluence l am tm = trace ("edges_confluence (dim = " ++ show vcount ++ ")") $ -- traceShow ("degs", degs) $
  matInfo "am" am $ matInfo "tm" tm $
  SMatrix.toList matconf'
    where
      vcount   = natToInt @n
      degs     = DMatrix.extract $ SV.toDense (sumWith Colonne identity am)
      !sumdeg   = VS.sum degs
      !tmam     = SMatrix.zipR 0 (,) am tm
      matconf' = seq sumdeg $ trace "edges_confluence.matconf'" $ SMatrix.imap
                 (\x y _v ->
                   if x < y
                      then
                        let
                            !deg_x = (VS.!) degs x - 1
                            !deg_y = (VS.!) degs y - 1

                            !tm' = {- trace ("tm': " ++ show (x, y)) $ -} SMatrix.imap
                              (\i j (am_ij, tm_ij) ->
                                 if (i == x && j == y) || ( i == y && j == x)
                                 then 0
                                 else
                                   if i == x && i /= y
                                   then am_ij / deg_x
                                   else
                                     if i == y && i /= x
                                     then am_ij / deg_y
                                     else tm_ij
                              ) tmam

                            v  = fromList (Proxy @n) [(fromIntegral y,1)]
                            v' = doProx l v tm'

                            prox_y_x_length = coeffSM 0 x v'
                            prox_y_x_infini = deg_x / (sumdeg - 2)

                        in -- traceShow ("ec",x,y,deg_x,sumdeg,prox_y_x_infini) $
                             (prox_y_x_length - prox_y_x_infini)
                           / (prox_y_x_length + prox_y_x_infini)

                      else 0
                 ) am

      doProx
        :: forall           n
         . KnownNat         n
        => Length
        -> VectorS     n
        -> TransitionMatrix n
        -> VectorS     n
      doProx l v' tm'' = case l <= 1 of
          True  -> panic "doProx"
          False -> foldl' (\v'' _-> SMatrix.mul v'' tm'') v' [1 .. (l :: Int) ]

sort_edges :: Int
           -> UnsortedEdges
           -> SortedEdges
sort_edges n = trace "sort_edges"
             . List.sortBy (\a b -> comparing third a b <> comparing xnpy a b)
             . List.filter (\(x,y,_) -> x < y)
  where
    third
      :: forall a b c
       . (a,b,c) -> c
    third (_,_,c) = c
    xnpy (x,y,_) = x*n+y


---------------------------------------------------------------
updateClustering :: Clustering Node
                 -> (Int -> Int -> Double)
                 -> Int -> Int -> Clustering Node
updateClustering c@(ClusteringIs parts idx currentScore _) f x y =
  let modX = fromMaybe 0 $ Dict.lookup x idx
      modY = fromMaybe 0 $ Dict.lookup y idx
   in case x == y || modX == modY of
    True  -> c -- do case x' or y' are Nothing
    False -> let c' = updateWith c f x modX y modY
              in case score c' >= currentScore of
                   True  -> c'
                   False -> c

updateWith :: Clustering Node
           -> (Int -> Int -> Double)
           -> Int -> Int
           -> Int -> Int
           -> Clustering Node
updateWith c@(ClusteringIs parts idx _ _) f x modX y modY = trace "updateWith" -- traceShow ("score", x,y,score') $
  ClusteringIs parts' idx' score' Part
    where
      parts' = Dict.filter (not . Set.null)
             $ Dict.alter (add y) modX
             $ Dict.alter (del y) modY parts

      add y Nothing   = Just (Set.singleton y)
      add y (Just y') = Just (Set.insert y y')

      del y Nothing   = Nothing
      del y (Just y') = Just (Set.delete y y')

      idx'         = Dict.alter (alter modX) y idx
      alter my _   = Just my

      px = Dict.elems parts'

      score' = getSum $
        foldMap (\s -> fold [ Sum (f x'' y'')
                            | x'' <- Set.toList s
                            , y'' <- Set.toList s
                            , x'' < y''
                            ]
                ) parts'

        -- Prelude.sum $ List.concatMap
        -- (\s -> [ f x'' y'' -- to Validate with BG
        --        | x'' <- Set.toList s
        --        , y'' <- Set.toList s
        --        , x'' < y''
        --        ]
        -- ) px

---------------------------------------------------------------
make_clust_part :: forall n
                 . KnownNat n
                => SortedEdges
                -> SimilarityMatrix n
                -> Clustering Node
make_clust_part se sm = trace "make_clust_part" $ foldl' (\c (e1,e2,_)
                                 -> {- trace ("foldl' step: " ++ show (e1, e2)) $ -} updateClustering c
                                    (\x y -> 2 * (coeffDM x y sm'))
                                    e1 e2
                               ) (ClusteringIs parts idx 0 Part) se
  where
    ns    = [0 .. natToInt @n - 1 ]
    parts = Dict.fromAscList (List.zip ns $ map Set.singleton ns)
    idx   = Dict.fromAscList (List.zip ns ns)
    sm'   = case sm of
      SimConf cm -> cm
      SimMod  mm -> mm

------------------------------
make_clust_over :: KnownNat n
                => SimilarityMatrix n
                -> StrictClustering   a
                -> OverlapsClustering a
make_clust_over = Prelude.undefined

---------------------------------------------------------------
-- | Some Utils

-- | The Main Trick: Graph must be finite, then n can be known 1 time at least
--  The same n is needed the size of the Graph and Matrixes and Vectors
--  Proof by compilation of the compatibilities of the calculations
--  Then Proof by tests is less needed at this specific step of the flow.
buildFiniteGraph :: Proxy n -> Graph a b -> FiniteGraph n a b
buildFiniteGraph _ g = FiniteGraph g

-- | To execute a function on finite Graph
-- the KnownNat n is computed one time only
withG :: (Show a, Show b)
      => Graph a b
      -> (forall n. KnownNat n => FiniteGraph n a b -> r)
      -> r
withG g f = reifyNat (fromIntegral $ length $ nodes g)
          $ \n -> f (buildFiniteGraph n g)

{-
random :: forall n. Proxy n -> IO (MatrixD n)
random _ = DMatrix.random

withM :: Integer
      -> (forall n. KnownNat n => MatrixD n -> r)
      -> r
withM n f = reifyNat (fromIntegral n)
          $ \n -> do
              m <- random n
              pure $ f m
-}


---------------------------
-- Some Sugar to be generalized: same functions for either Dense or
-- sparse Matrix can be unified thanks to the Type instances.
class Symmetry a where
  symmetry :: a -> a

instance KnownNat n => Symmetry (MatrixD n) where
  symmetry = symmetryD

instance KnownNat n => Symmetry (MatrixS n) where
  symmetry = symmetryS

symmetryD :: KnownNat n => MatrixD n -> MatrixD n
symmetryD m = DMatrix.imapL (\(x, y) v -> if x < y then v else coeffDM y x m) m

symmetryS :: KnownNat n => MatrixS n -> MatrixS n
symmetryS m = SMatrix.imap (\x y v -> if x < y then v else coeffSM y x m) m

---------------------------
class KnownNat n => Vector n a where
  fromList :: KnownNat n => Proxy n -> [(Integer,Double)] -> a

instance KnownNat n => Vector n (VectorS n) where
  fromList = vectorFromListS

instance KnownNat n => Vector n (VectorD n) where
  fromList = vectorFromListD

vectorFromListS :: KnownNat n => Proxy n -> [(Integer, Double)] -> VectorS n
vectorFromListS pn ns
   | all (\(i, v) -> i >= 0 && i < n) ns = SMatrix.fromList $ map (\(i,v) -> (0, fromInteger i,v)) ns
   | otherwise                           = panic "error vectorFromListS"
     where
       n = natVal pn

vectorFromListD :: KnownNat n => Proxy n -> [(Integer, Double)] -> VectorD n
vectorFromListD pn ns = SMatrix.densify $ vectorFromListS pn ns

------------------------------
data Direction = Ligne | Colonne

sumWith
  :: (UV.Unbox a, UV.Unbox t, KnownNat n, Num t, HasCallStack, Show t)
  => Direction -> (t -> a) -> SMatrix.Matrix n n t -> SV.V n a
sumWith dir f m = SV.map f $
  case dir of
    Colonne -> SV.sum $ SMatrix.getCols m
    Ligne   -> SV.sum $ SMatrix.getRows m

--------------------------------------------

natToInt :: forall (n :: Nat). KnownNat n => Int
natToInt = fromIntegral $ natVal (Proxy :: Proxy n)

coeffSM
  :: (Num a, UV.Unbox a, Show a)
  => Int -> Int -> SMatrix.Matrix n p a -> a
coeffSM i j m = SMatrix.extractCol m j SV.! i

coeffDM
  :: (Num a, KnownNat n, KnownNat p)
  => Int -> Int -> DMatrix.L n p -> Double
coeffDM i j m = m `DMatrix.at` (i, j)
